%% Core-Set-Soft-Competitive-Learning (with subsampling)
%
% -- not for linear kernel -- due to distance optimization
%
% Input:
%       mD        - N x D   -- data matrix, with N samples and D dimensions
%       (should be normalized to N(0,1) 
%       mD=bsxfun(@rdivide,bsxfun(@minus,mD,mean(mD)),(eps+std(mD)));
%       K         - integer -- number of expected clusters
%       epsilon   - double  -- coreset tolerance
%       iCyles    - integer -- number of maximal cycles
% Output:
%       mAlpha    - alpha matrix for the cluster centers with respect to the kernel
%
%
% For labeled data with numerical labels stored in L you can do the post-labeling as
% dAcc = 0; vUnique = unique(vWinners); for(k=1:length(vUnique)),vField = vWinners==vUnique(k); vPosts(k) = mode(L(vField)); dAcc=dAcc+sum(L(vField) == vPosts(k));end
% dAcc/length(L)
% The neighborhood cooperation takes place by means of the distances to the core-set points
%% 

function [mAlpha,vWinners,vConvergenceTest] = core_ng(mD,K,epsilon,iCycles,fKernel,varargin)
addpath argmaxmin;
addpath qpc;
addpath kernel;
addpath helper;
if(~(ismatrix(mD) && isnumeric(K) && K>0 && isreal(epsilon) && epsilon>0 && epsilon<=1 && isnumeric(iCycles) && iCycles>=1))
	error('Invalid input parameters');
end

% initial codebook
[dlen]   = size(mD,1);
lambda0  = K/2; 
alpha0   = 0.15;

vFull         = [1:dlen];
dSamplingRate = 1;
vFieldIndices = zeros(dlen,1);
mAlpha        = zeros(K,dlen);

% random init
vWinners      = ceil(rand(1,dlen)*K);
vLabels       = vWinners;

% the diags could be provided from outside (e.g. if constant w.r.t. kernel function)
fKernelPtrRangeF    = @(X,Y)fKernel(mD(X,:),mD(Y,:));
vDiags              = zeros(1,dlen);
for(k=1:dlen)
    vDiags(k)       = fKernel(mD(k,:),mD(k,:));
end

% neighborhood influence
lambda  = lambda0 * (0.001/lambda0).^([0:(iCycles-1)]/iCycles);

% learn rate
alpha  = alpha0 * (0.005/alpha0).^([0:(iCycles-1)]/iCycles);

vRanks     = [0:(K-1)];

mAlphaBack = mAlpha; % prior alpha matrix to keep track of SCL updates

% distance Neurons to data - first round
vRandSample= randsample(dlen,min(max(100,ceil(dSamplingRate*dlen)),dlen));
vWinners   = vWinners(vRandSample);
vRandSampleConvTest  = randsample(dlen,ceil(0.1*dlen));
vConvergenceTest = zeros(iCycles,1);
vCurrentCoreSetIndices = false(1,dlen);
k    = 1;   
last = 0;
vLastTest       = vWinners;
while(k < iCycles && any(vLastTest ~= last))
    if(0)        
        if(dSamplingRate==1) % do a subsampling if full set is used
                fKDistPtr    = @(X,Y)((bsxfun(@minus,bsxfun(@plus,vDiags(vRandSampleConvTest(X(:))),vDiags(vRandSampleConvTest(Y(:)))'),2*fKernelPtrRangeF(vRandSampleConvTest(X),vRandSampleConvTest(Y))')));                
                [DualCostFuncVal]   = determine_cost_function_dual_approx( fKDistPtr, vWinners(vRandSampleConvTest));        
        else
                fKDistPtr    = @(X,Y)((bsxfun(@minus,bsxfun(@plus,vDiags(vRandSample(X(:))),vDiags(vRandSample(Y(:)))'),2*fKernelPtrRangeF(vRandSample(X),vRandSample(Y))')));           
                [DualCostFuncVal]   = determine_cost_function_dual_approx( fKDistPtr, vWinners);        
        end
        vConvergenceTest(k) = DualCostFuncVal;%sum(vWinners~=last);
        if(k>10)
            dMedianQError       = median(vConvergenceTest(1:k-1));
            if(abs(dMedianQError/(dMedianQError+DualCostFuncVal)-0.5) < 1e-3) %~any(vWinners~=last))
                    'converged'
                    break;
            end
        end
    end

    % generate new centers
    mAlpha(:,vCurrentCoreSetIndices) = 0;
    vCurrentCoreSetIndices(:) = false;

    for(c=1:K) % for each cluster over all ranks    
        % determine MEB of current prototype field
        vCurrentFieldIndices = vRandSample((vWinners==c));         
        if(numel(vCurrentFieldIndices)>1)
            DReduced   = mD(vCurrentFieldIndices,:);
            fKernelPtr = @(X,Y)fKernel(DReduced(X,:),DReduced(Y,:));
            [S,R,vAlpha]  = core_set(fKernelPtr,epsilon,vDiags(vCurrentFieldIndices));
            mAlpha(c,vCurrentFieldIndices(S))   = vAlpha'; 
            vCurrentCoreSetIndices(vCurrentFieldIndices(S)) = true;
        elseif (numel(vCurrentFieldIndices)>0)
            mAlpha(c,vCurrentFieldIndices)   = 1; 
            vCurrentCoreSetIndices(vCurrentFieldIndices) = true;
        end
        vFieldIndices(vCurrentFieldIndices) = c;
    end
    

    vCandiatesForCoreSetIndices = find(vCurrentCoreSetIndices==true);
    iNrCandidates               = length(vCandiatesForCoreSetIndices);
    vRand                       = randsample(iNrCandidates,min(max(100,ceil(0.1*iNrCandidates)),iNrCandidates)); 
    vCandiatesForCoreSetIndices = vCandiatesForCoreSetIndices(vRand);                

    iNumberSamples     = length(vCandiatesForCoreSetIndices);
    mDistances2Neurons = zeros(K,iNumberSamples);
    for(c=1:K)
        vAlpha          = mAlpha(c,:);
        S               = find(vAlpha);
        vAlpha          = vAlpha(S);        
        mK              = fKernelPtrRangeF(S,vCandiatesForCoreSetIndices);
        mKSS            = fKernelPtrRangeF(S,S);
        vDistances      = -2 * vAlpha * mK + vAlpha * mKSS * vAlpha'; % skip sqrt
        mDistances2Neurons(c,:) = vDistances;
    end            

    [~,mI]          = sort(mDistances2Neurons);
    ranking         = zeros(iNumberSamples,K);
    for(i=1:iNumberSamples)
        ranking(i,mI(:,i))   = vRanks;             
    end

    % update
    mTemp  = alpha(k) * exp(-ranking./lambda(k));
    vMean  = mean(mTemp);
    mAlpha = mAlpha - bsxfun(@times, mAlphaBack,vMean'); 
    mAlpha(:,vCandiatesForCoreSetIndices) = mAlpha(:,vCandiatesForCoreSetIndices) + mTemp';

    mAlpha(mAlpha<0) = 1e-10;                                           % avoid underflow            
    mAlpha           = bsxfun(@rdivide,mAlpha,1e-10+sum(mAlpha,2));     % normalize
        
    % get new receptive fields
    if(dSamplingRate~=1)
        vRandSampleBack = vRandSample; % store the last rand selection
        vRandSample     = randsample(dlen,min(max(100,ceil(dSamplingRate*dlen)),dlen));
    end

    S          = find(any(mAlpha));
    mK         = fKernelPtrRangeF(S,vRandSample);
    mKSS       = fKernelPtrRangeF(S,S);
    mAlphaShort= mAlpha(:,S);
    mDistances = bsxfun(@plus,bsxfun(@minus,vDiags(vRandSample),2 *mAlphaShort * mK), diag(mAlphaShort * mKSS * mAlphaShort'));
    mDistances(mDistances<0) = 0;    
    last       = vWinners; % all points used
    vWinners   = argmin(mDistances,1);

    % for convergence test
    if(dSamplingRate == 1)        
        vLastTest       = vWinners;
    else
        % get the distances for the last randsample selection        
        mK         = fKernelPtrRangeF(S,vRandSampleBack);
        mKSS       = fKernelPtrRangeF(S,S);
        mAlphaShort= mAlpha(:,S);
        mDistances = bsxfun(@plus,bsxfun(@minus,vDiags(vRandSampleBack),2 *mAlphaShort * mK), diag(mAlphaShort * mKSS * mAlphaShort'));
        mDistances(mDistances<0) = 0;    
        vLastTest  =  argmin(mDistances,1);
    end
    
    if(mod(k,1)==0)
        fprintf('* %f *',k/iCycles);
    end
    vCurrentCoreSetIndices(:) = false;
    vCurrentCoreSetIndices(S) = true;    % define all non-vanishing entries as core set points
    mAlphaBack                = mAlpha;  % copy current alpha matrix         
    k = k + 1;
end  

% distance Neurons to data - full
S          = find(vCurrentCoreSetIndices);
mKSS       = fKernelPtrRangeF(S,S);
iStepSize  = 10000;
iStart     = 1;
mAlpha(mAlpha<1e-5) = 0;
mAlphaShort= mAlpha(:,vCurrentCoreSetIndices);
C          = diag(mAlphaShort * mKSS * mAlphaShort');
mDistances = zeros(K,dlen);
while(iStart<dlen)
    iEnd       = min(iStart + iStepSize,dlen);
    mK         = fKernelPtrRangeF(S,vFull(iStart:iEnd));
    mDistances(:,[iStart:iEnd]) = bsxfun(@plus,bsxfun(@minus,vDiags(vFull(iStart:iEnd)),2 *mAlphaShort * mK), C);
    iStart     = iEnd + 1;
end
mDistances(mDistances<0) = 0;
vWinners   = argmin(mDistances,1);
