%%%%%%%%%%%%5
%
% Algorithm to calculate the core set of a set of points
% with probabilistic sampling (see JMLR, Tsang & Kwok, 2005)
% (in an SVDD manner)
%
% 
% IN:
%
% - fKPtr   - pointer to a kernel function e.g. fKPtr = @(X,Y)K(X,Y) - for a kernel matrix K
% - epsilon - eps of the minimum enclosing ball problem (MEB)
% - vDiags  - pre-calculated K(x,x) values for each x
%
% OUT:
% - S - set of core set points 
% - R - radius of the core set
% - vAlpha - loadings of the QP solution for S
%
% Note:
%
% In contrast to std core-set / MEB solution the algorithm has an internal
% outlier tolerance of 10% - meaning that the MEB is ok also if upto 10% of the points are not in the epsilon-MEB
% otherwise the algorithm chooses as the next support point one which is
% out of epsilon-MEB and is closest to (the typical distance of) this set.
% In this way we avoid that outlier (potential alternative cluster points)
% are included in the set.
% - The original argmax strategies is out-commented
% 
% implemented 2011 by Frank-M. Schleif <fschleif@techfak.uni-bielefeld.de>
%
%% 
function [S,R,vAlpha] = core_set(fKPtr,epsilon,vDiags) % using only the kernel

% check input arguments
if(~exist('epsilon','var') || epsilon <= 0)% || epsilon > 1)
    epsilon = 0.1;   
end

% provide alternative quadratic-progr. solver if available
if(exist('qpas','file')==3) % spm-qpc newcastle library available
    fPtrMinimalEnclosingBall = @(X,x0)MinimalEnclosingBallSPM(X,[]); % x0 is a dummy - warm start is not supported by spm-qpc
else
    fPtrMinimalEnclosingBall = @(X,x0)MinimalEnclosingBall(X,x0);
end

vDiags = vDiags(:);
N      = numel(vDiags);
iProbSampleSize = 59;
%% init 
fKDistPtr    = @(X,Y)((bsxfun(@minus,bsxfun(@plus,vDiags(X(:)),vDiags(Y(:))'),2*fKPtr(X,Y)')));
dTolerance   = (1+epsilon).^2; % for the epsilon problem
dOutlierPercentage = 0.1;  % 1% - 0.01       
SFull        = [1:N]; 
bSimple      = false;
if(bSimple)
    S            = randsample(N,2)';                                              % choose 2 random points (could be done more clever - but the cluster maybe a mix - so be careful)
else
    % identify raw center of the cloud
    vRandSampling= randsample(length(SFull),min(length(SFull),iProbSampleSize));               % some random samples    
    %Dists        = squareform(pdist(SFull(vRandSampling)',fKDistPtr));           % thats slower
    Dists        = fKDistPtr(SFull(vRandSampling),SFull(vRandSampling));
    dMeanDist    = mean(Dists(triu(Dists)~=0));
    vTypicalPairs= argmin(abs(Dists-dMeanDist));
    [vCol,vRow]  = ind2sub(size(Dists),vTypicalPairs);
    S(1)         = vCol(1); % first pt (median of both should be somehow in the center)
    S(2)         = vRow(1); % snd pt 
end
SFull        = SFull(SFull~=S(1) & SFull~=S(2));                              % remove new added entry from SFull
vAlpha       = 0.5*ones(2,1);                                                 % default loadings in case of 2 points (initial center)
mKSS         = fKPtr(S,S);                                                    % kernel matrix
R            = (vAlpha' * diag(mKSS) -vAlpha' * mKSS * vAlpha);               % thats R^2 - initial radius
if(isempty(SFull))
    return;
end

%% core set calculation - inital step
iProbSampleSize = 59;%length(SFull);
vRandSampling= randsample(length(SFull),min(length(SFull),iProbSampleSize));

mK           = fKPtr(S,SFull(vRandSampling));                                                % kernel of core points to all out of radius points
mKSS         = fKPtr(S,S);      % SFull is without S    
vDistances   = (vDiags(SFull(vRandSampling))' -2 * vAlpha' * mK + vAlpha' * mKSS * vAlpha);   % get distances for MEB
dMed         = median(vDistances);                                            % median distance of the current core set to all other - estimate used later on


%% loop  until MEB is found
iCounter     = 2;
while(~isempty(SFull))      
     vOutOfRadius     = find(vDistances > dTolerance * R);                    % identify points out of MEB
     if(length(vOutOfRadius)<=dOutlierPercentage*length(vDistances))                          % if there are not to many it is still ok
         break;
     end     
     %%I               = argmax(vDistances(vOutOfRadius),2);                   % largest point away - original core-set heuristic
     %%I                = argmin(vDistances(vOutOfRadius),2);                   % closest point away - original core-kmeans heuristic
     I                = argmin(abs(vDistances(vOutOfRadius)-dMed),2);         % point with the most typical distance to all other points (hope to keep outliers out)
     iNewCoreIndex    = SFull(vRandSampling(vOutOfRadius(I)));                % identify new coreset point
     iCounter         = iCounter+1;
     S                = [S, iNewCoreIndex];                                   % add to core-set
     SFull            = SFull(SFull~=iNewCoreIndex);                          % remove new added entry from SFull and all which are in the circle
     mKSSOld          = mKSS;
     iEntries         = length(S);
     mKSS             = zeros(iEntries);                                      % update small kernel for core-set points (a^2), 2ab used in MEB optimization
     mKSS([1:iEntries-1],[1:iEntries-1]) = mKSSOld;
     mKSS(iEntries,:) = fKPtr(S,iNewCoreIndex);                               % calculate the kernel values for the new entry
     mKSS(:,iEntries) = mKSS(iEntries,:);
     vAlpha0          = [vAlpha; 0.5]';
     vAlpha0          = vAlpha0./sum(vAlpha0);    
     [R,vAlpha]       = fPtrMinimalEnclosingBall(mKSS,vAlpha0);                                     % update radius (with warm start alpha if available
     iProbSampleSize  = 59;%length(SFull);
     vRandSampling    = randsample(N-iCounter,min(N-iCounter,iProbSampleSize)); 
     mK               = fKPtr(S,SFull(vRandSampling));                                              % kernel of core points to all out of radius points
     vDistances       = (vDiags(SFull(vRandSampling))' -2 * vAlpha' * mK + vAlpha' * mKSS * vAlpha); % calculate new distances
     vDistances(vDistances<0) = 0;
     dMed         = median(vDistances);                                            % median distance of the current core set to all other - estimate used later on
end
% length(vAlpha)
% remove entries which vanished from the optimization - may happen 
S      = S(vAlpha>eps);
vAlpha = vAlpha(vAlpha>eps);
R      = sqrt(R); % return only R not R^2

function [R,vAlpha] = MinimalEnclosingBall(mK,vAlpha0)
    N = size(mK,1);
    options = struct('Diagnostics','off','Algorithm','active-set');
    vAlpha = quadprog(2*mK,-diag(mK),[],[], ones(1,N),1, zeros(N,1),ones(N,1),vAlpha0,options); % here we use the dual formulation including a kernel (here linear, euclidean)
    R = (vAlpha' * diag(mK) -vAlpha' * mK * vAlpha);

function [R,vAlpha] = MinimalEnclosingBallSPM(mK,x0)
    N = size(mK,1);
    vAlpha = qpas(2*mK,-diag(mK),[],[],ones(1,N),1,zeros(N,1),ones(N,1)); % spm-qpc (newcastle) library
    R = (vAlpha' * diag(mK) -vAlpha' * mK * vAlpha);
    
    