function [vModelResults,vClassifyResults,vT] = crossvalidation_cng(rule, SDL, v, opt)
%CROSSVAL K-fold cross validation.
%   [C, POST] = CROSSVAL(RULE, X, K) returns the leave-one-out
%   cross-validation classification estimates of the observations
%   described by the feature matrix X and the class index vector K
%   given the name of the classifier RULE. RULE must be a string
%   containing the name of a CLASSIFIER object such as one of the
%   descendents of the class 'classifier'. X is an n by p matrix where
%   n is the number of observations and p is the number of
%   variables. K is a length n vector of class indeces in which each
%   element gives the group to which the corresponding row in X
%   belongs. C contains the estimated classificiations in a length n
%   class index vector and POST is a n by g matrix of posterior
%   probabilities.
%
%   [C, POST, V] = CROSSVAL(RULE, X, K, NSETS) where NSETS is a scalar
%   value no more than n perfoms NSETS-fold cross-validation with
%   randomly chosen sets of approximately equal size. The return
%   argument V is a vector of indeces indicating the cross-validation
%   group to which each row of X and K were assigned. If NSETS is a
%   vector of indeces, it is used as the cross-validition sets
%   instead. Empty sets are not permitted in NSETS.
%
%   CROSSVAL(RULE, X, K, OPTS) allows optional arguments to be passed
%   in the fields of OPTS. The only field that is used by CROSSVAL is
%   NSETS. Additionally, OPTS is passed to the CLASSIFIER constructor
%   as well as the function CLASSIFY for the appropriate CLASSIFIER
%   class so that options may be assigned to those funcitons as well.

%   Copyright (C) Michael Kiefte 1998.

%   $Log$
%   call: 
%   [vModelResults,vClassifyResults] =
%   crossvalidation_clustering(@kngas_ny_cv, oSDL, 5, {3,10,0.1,3/2,0.05,'nystroem',0.1,'sparse_beta',10,'sparse_sigma',0.001,'few_memory',0} )
%
X = SDL.data;
k = SDL.labels;

error(nargchk(3, 4, nargin))
[n p] = size(X);
if nargin < 3
  v = [];
end

if nargin < 4
  opt = [];
end

if isempty(v)
  V = 1:n;
  v = n;
elseif length(v) == 1
  if ~isa(v, 'double') | ~isreal(v) | v <= 0 | round(v) ~= v
    error('V must be a positive, non-zero integer.')
  elseif v > n
    error('V must be less than the number of observations.')
  elseif v == n
    V = 1:n;
  else
    w = v*fix(n/v);
    r = randperm(v);
    V(randperm(n)) = [reshape(repmat(1:v, w/v, 1), w, 1); ...
		      r(1:n-w)'];
  end
else
  if ~isa(v, 'double') | ~isreal(v) | prod(size(v)) ~= length(v) | ...
	any(round(v) ~= v | v <= 0 | isinf(v))
    error(['Cross-validation set indices V must be a vector of' ...
	   ' positive, finite, non-zero integers.'])
  elseif length(v) ~= n
    error(['Length of set indeces V must be same as number of' ...
	   ' observations in X and K.'])
  elseif ~all(sum(sparse(1:n, v, 1)))
    error('Set indeces V may not have empty groups.')
  end
  
  V = v;
  v = max(V);
end
  
c = zeros(n, 1);
vModelResults    = [];
vClassifyResults = [];
vT = []; vM = [];
for i = 1:v
  s = warning;
  warning off
  SDLCurrent = SDL;
  SDLCurrent.data   = SDLCurrent.data(V ~= i, :);
  SDLCurrent.labels = SDLCurrent.labels(V ~= i, :);
  dT = cputime;
  [mAlpha,vWinners]    = feval(rule, SDLCurrent.data, opt{:});
  dTRuntime = cputime-dT;
  % post-label clusters
  dAccTrain = 0; vUnique = unique(vWinners); % empty fields are removed
  L         = SDLCurrent.labels;
  for(k=1:length(vUnique)),
      vField = vWinners==vUnique(k); 
      vPosts(k) = mode(L(vField)); 
      dAccTrain = dAccTrain+sum(L(vField) == vPosts(k));
  end   
  dAccTrain = dAccTrain/length(vWinners);
  oModel.mAlpha           = mAlpha(vUnique,:);
  oModel.vWinners         = vWinners;
  oModel.PrototypeLabels  = vPosts;
  oModel.dAccTrain        = dAccTrain;
  oModel.oTrain           = SDLCurrent;
  oModel.options          = {opt};  
  oModel.vLabelMapping    = unique(L);  
  oModel.dSparsity        = nnz(oModel.mAlpha)./numel(oModel.mAlpha);
  vT(i) = dTRuntime;

  SDLTest = SDL;
  SDLTest.data   = SDLTest.data(V == i, :);
  SDLTest.labels = SDLTest.labels(V == i, :);  
 
  [dAcc,mClasses,vLabelsAssigned,vAssignments] = LabelNewData(oModel,SDLTest);
  
  DualCostFuncVal = 0;
  meanDist        = 0;
  if(length(oModel.options{1})>3)
    fKernel         = oModel.options{1}{4};
  else
        'assume linear kernel'
        fKernel         = @(X,Y)X*Y';
  end
  if(1)
    dlen             = numel( SDLTest.labels);
    vRand            = randsample(dlen,min(1000,dlen));
    dlen             = numel(vRand);
    vDiags           = zeros(1,dlen);
    fKernelPtrRangeF = @(X,Y)fKernel(SDLTest.data(vRand(X),:),SDLTest.data(vRand(Y),:));
	for(k=1:dlen)
        vDiags(k)    = fKernelPtrRangeF(k,k);
	end    
    fKDistPtr        = @(X,Y)((bsxfun(@minus,(vDiags(X)+vDiags(Y)),2*fKernelPtrRangeF(X,Y))));
    'start dual calculation'
    [DualCostFuncVal, meanDist]  = determine_cost_function_dual( fKDistPtr, vAssignments(vRand) );
  end  
  
  vClassifyResults        = [vClassifyResults dAcc];
  oModel.DualCostFuncVal  = DualCostFuncVal;
  oModel.meanDist         = meanDist;
  vModelResults           = [vModelResults oModel];  
  fprintf('*\n');
  warning(s);  
end









